/*++++++++++++++
.IDENTIFICATION util.c
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT    Unix, CDS Catalogues
.KEYWORDS       Table, FITS
.VERSION  1.0   15-May-1991 (anafile)
.VERSION  4.0   28-Feb-1994 New package
.VERSION  4.1   01-Oct-1996: epochs
.VERSION  4.2   10-Jun-1997: domain
.VERSION  4.3   02-Sep-1997: Added strdup2 strdup1 strdupv
.VERSION  4.4   06-Apr-1998: Added strindex
.VERSION  4.5   18-Sep-1998: Options 'F' (FITS) and 'T' (FITS with Time) in date
.VERSION  4.6   20-Apr-2000: Option in epochs -- 1 to get any word following
.VERSION  4.7   29-May-2000: Eplit in 2 parts  strut.c util.c
.VERSION  4.71  07-Mar-2001: XML2FITS Context
.VERSION  4.72  19-Apr-2001: Added dateset
.VERSION  4.72  13-Mar-2002: Accept epoch written as   at "Column-Epoch"
.VERSION  4.73  16-Apr-2002: Just a better parser for Epoch/Equinox
.VERSION  4.74  21-Sep-2002: epochs() returns also the words 'Equinox' 
				and 'Epoch'
.VERSION  4.75  26-May-2006: use isGraph (any 8-bit char is graph)
.VERSION  4.77  02-Jun-2007: Cosmetics (-Wall)
.VERSION  4.78  11-Jul-2007: epoch may be followed by 'is'
.VERSION  4.79  21-mai-2010: bandGHz to interpret wavelength/freq/energy
.VERSION  4.80  02-Sep-2010: Accept 'AD' (Anno Dimini) for Epoch / Equinox
.COMMENTS       Small C String Utilities for anafile package
---------------*/

#ifndef BZERO
#define BZERO 0
#endif

#ifndef DEBUG
#define DEBUG 0
#endif

#ifdef XML2FITS
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
int strmatch(), strloc();
#define issign(c)      (((c) == '+') || ((c) == '-'))
#else
#include <catstd.h>
#endif
#include <string.h>	/* Standard string manipulations */
#include <time.h>
#include <units.h>

#if DEBUG
#define RETURN(x) 	{ printf("%d\n", x); recursivity--;\
		for (i=0; i<recursivity;i++) putchar(' '); \
		return(x); } 					/* Debugging */
static char inbyte;
static int i, recursivity;
#else
#define RETURN(x)	return(x)
#endif

#define ____    000     /* NULL character       */
#define MAX(x,y)        ((x) >= (y) ? (x) : (y))
#define MIN(x,y)        ((x) >= (y) ? (y) : (x))
#define ITEMS(a)        sizeof(a)/sizeof(a[0])

/* Problem with 8-bit chars */
#define isGraph(x)	(isgraph(x) || ((x)&0x80))

static char sep[] = " ,()[];\t";	/* Separator characters */

/*==================================================================
		String utilities
 *==================================================================*/
char *rule(c1)
/*++++++++++++++++
.PURPOSE  Defines a 80-byte char rule made of c1
.RETURNS  The 80-byte character string
.REMARKS  '\' generates TeX, '<' the HTML
-----------------*/
  int c1;       /* IN: Character for repetition   */
{
  static char line[81];
  int  i;
  
    if(c1 == '\\') return("\\hr");
    if (line[0] != c1) for (i=0; i<sizeof(line);i++) line[i] = c1;
    line[sizeof(line)-1] = 0;
    return(line);
}

char *strclone(s)
/*++++++++++++++++
.PURPOSE  Duplicate (clone) a string
.RETURNS  The (newly allocated) string
-----------------*/
  const char *s; /* IN: String to clone */
{
  char *c;
    if ((c = malloc(strlen(s)+1))) strcpy(c, s);
    return(c);
}

int strnline(s, len)
/*++++++++++++++++
.PURPOSE  Compute length of line (remove trailing spaces)
.RETURNS  The length of the trimmed line
-----------------*/
  char *s;	/* IN: String to scan */
  int len;	/* IN: Valid length of s */
{
  char *p;
    for (p=s+len; p>s;) {
    	--p;
    	if (isspace(*p)) continue;
	p++; break;
    }
    return(p-s);
}

int strline(s)
/*++++++++++++++++
.PURPOSE  Compute length of line (remove trailing spaces)
.RETURNS  The length of the trimmed line
-----------------*/
  char *s;	/* IN: String to scan */
{
  return(strnline(s, strlen(s)));
}

int setset1e(in, es, set256)
/*++++++++++++++++
.PURPOSE  Set to 1 an array of 256 bytes for chars appearing in in.
.RETURNS  Length of interpreted input string
.REMARKS  A-Z for range  A to Z
-----------------*/
  char *in;       /* IN: String to interpret */
  char *es;       /* IN: End of string in    */
  char *set256;   /* OUT: Array where bytes appearing in in are set to 1 */
{
  char *p, u, c; int b;

    --es ;	/* Last byte of in */
    for (p = in, c=0177; p <= es; p++) {
	if ((*p == '-') && (p < es)) {	/* A range ? */
	    u = p[1] ;
	    if (u > c) {	/* Yes, range*/
	    	for (b=c; ++b <= u; ) set256[b] = 1;
	    	p++;
	    	c = 0177;
	    	continue;
	    }
	}
	c = *p ;
	set256[c&0xff] = 1; 
    }
    return (p-in);
}

int setset1(in, set256)
/*++++++++++++++++
.PURPOSE  Set to 1 an array of 256 bytes for chars appearing in in.
.RETURNS  Length of interpreted input string
.REMARKS  \octal or \f \b \r \n, A-Z for range  A to Z
-----------------*/
  char *in;       /* IN: String to interpret */
  char *set256;   /* OUT: Array where bytes appearing in in are set to 1 */
{
    return(setset1e(in, in+strlen(in), set256)) ;
}

/*==================================================================*/
char *edchar(c)
/*++++++++++++++++
.PURPOSE  Edit a char (ascii or Octal)
.RETURNS  The edited string
-----------------*/
  int  c;         /* IN: Char to look for */
{
  static char buf[6];

    if(isprint(c)) buf[0] = c, buf[1] = 0; 
    else sprintf(buf, "%02o", c&0xff);
    return(buf);
}

/*===========================================================================*/
int strindex(s1, s2)
/*++++++++++++++++
.PURPOSE  Locate substring
.RETURNS  Index of substring
-----------------*/
  char *s1;     /* IN: Main string 	*/
  char *s2;     /* IN: Substring 	*/
{
  char *s; int len ;
    len = strlen(s2) ;
    if (len == 0) return(0) ;
    for ( s = s1; *s; s++) {
	if (*s != *s2) continue ;
	if (strncmp(s, s2, len) == 0) break ;
    }
    return(s-s1);
}

int strsindex(s1, s2)
/*++++++++++++++++
.PURPOSE  Locate substring, spaces can be squeezed
.RETURNS  Index of substring / length of s1 if not found
-----------------*/
  char *s1;     /* IN: Main string, blanks&spaces squeezed	*/
  char *s2;     /* IN: Substring -- without consecutive spaces 	*/
{
  char *b, *p, *s; 
    for ( s = s1; *s; s++) {
	for (p=s, b=s2; *b; b++) {
	    if (isspace(*b)) {
		if (!isspace(*p)) break ;
		while (isspace(*p)) p++ ;
	    }
	    else if (*b == *p) p++ ;
	    else break ;
	}
	if (!*b) break ;	/* Found... */
    }
    return(s-s1);
}

/*===========================================================================*/
int studiff(s1,s2)
/*++++++++++++++++++++++++
.PURPOSE Compare (or compute difference of) two strings, case insensitive.
.RETURNS A positive value if s1 > s2, zero if strings are identical,
	a negative value if s1 < s2.
-----------*/
  char *s1;	/* IN: address of first string 	*/
  char *s2;	/* IN: address of 2nd string  	*/
{
  char 	*p, *q;
  char 	cp, cq;
    for (p=s1, q=s2; ; p++, q++) {
  	cp = toupper(*p), cq = toupper(*q);
	if (cp != cq)	break;
    	if (!cp)	break;
    }
    return(cp - cq);
}

int stundiff(s1,s2, len)
/*++++++++++++++++++++++++
.PURPOSE Compare (or compute difference of) two strings, case insensitive.
.RETURNS A positive value if s1 > s2, zero if strings are identical,
	a negative value if s1 < s2.
-----------*/
  char *s1;	/* IN: address of first string 	*/
  char *s2;	/* IN: address of 2nd string  	*/
  int  len; 	/* IN: Max length to compare	*/
{
  char 	*p, *q, *e;
  char 	cp=0, cq=0;
    for (p=s1, e=s1+len, q=s2; p<e ; p++, q++) {
  	cp = toupper(*p), cq = toupper(*q);
	if (cp != cq)	break;
    	if (!cp)	return(0) ;
    }
    return(p == e ? 0 : cp - cq);
}

/*==================================================================*/
static int streq(s1, s2)
/*++++++++++++++++
.PURPOSE  Function which returns 1 if strings are equal
.RETURNS  0 (no match) / 1 (match)
-----------------*/
  char *s1;	/* IN: First string  */
  char *s2;	/* IN: Second string */
{
    return(!strcmp(s1,s2));
}

int match_word(line, filename, sep)
/*++++++++++++++++
.PURPOSE  Check is "filename" is a word in "line"
.RETURNS  0 = not in Line
.REMARKS  line contains file templates separated by blanks or commas.
	When sep is 0, we assume a STRICT comparison (no wild char)
-----------------*/
  char *line;	/* IN: Line containing file templates separated by blanks */
  char *filename;	/* IN: File name to check */
  int  sep;		/* IN: Separator char */
{
  char *p=(char *)0, *e, *t, x;
  int (*cmp)();

#if DEBUG
    printf("....match_word(\"%s\", \"%s\", '%c')",
    	line, filename?filename:"(NULL)", sep);
#endif

    if (!line) 	/* Match is OK if both are NULL */
    	return (! filename);
    
    	/* If a sep (e.g. ':') exists in the line, 
    	   keep only what follows this separator : 
	   Otherwise, we assume a STRICT comparison
     	*/
    t = line;
    cmp = streq;
    if (sep) {
    	t += strloc(t, sep);
	cmp = strmatch;
    	if (*t) t++; 
    }
    e = t + strlen(t);
    x = 0;
    while (t < e) {
	while (isspace(*t)) t++;
	if (*t == ',') t++;
	if (*t == ':') t++;
	while (isspace(*t)) t++;
	for (p=t; isGraph(*p) && (*p != ',') && (*p != ':'); p++) ; 
	x = *p; *p = 0;		/* Terminate string */
	if ((*cmp)(t, filename)) break;
	*p = x; t = p;		/* Restore  string  */
    }
#if DEBUG
    printf("=> %d\n", t<e);
#endif
    if (x) *p = x;		/* Restore  string  */
    return(t < e);
}

/*==================================================================*/
int match_rule(s)
/*++++
.PURPOSE Check wether s represents a rule separation 
		(at least 20 identical chars)
.RETURNS not-zero = match / 0 = mismatch
-------*/
  char *s;        	/* IN: Starting address */
{
  char *p;
    p = s;
    if (ispunct(*s)) while(*p == *s) p++;
    if ((p-s) < 20) return(0);
    if (!*p) return(1);
    return(isspace(*p));
}

/*===========================================================================*/
int match_num(s)
/*+++++
.PURPOSE Match a number (double floating), also in hexa form (0x...)
.RETURNS Number of bytes matched
.REMARKS 
-------*/
  char *s;        	/* IN: Starting address */
{
  char 	x, *p;
  
    for (p=s; *p == ' '; p++) ;	/* Skip leading blanks */
    if (issign(*p)) p++;

    while(isdigit(*p)) p++;
    if (*p == '.') {		/* We've a decimal number */
	p++;
	while(isdigit(*p)) p++;
    }
    if (isalpha(*p)) {
	x = toupper(*p);
	if ((x == 'E') || (x == 'D')) {
	    p++;
	    if (issign(*p)) p++;
	    while (isdigit(*p)) p++;
	}
    }
    return (p-s);
}

 /*==================================================================*/
int ebcdic(c)
/*++++++++++++++++
.PURPOSE  Tell if non-ascii character can be EBCDIC
.RETURNS  ASCII equivalent
-----------------*/
  int  c;         /* IN: Character to test */
{
  static char ebc_to_asc[256] = {
	  00,  01,  02,  03,____,0011,____,0177, /* 0. */
	____,____,____,0013,0014,0015,0016,0017, /* 0. */
	 020,0021,0022,0023,____,____,0010,____, /* 1. */
	 030,0031,____,____,0034,0035,0036,0037, /* 1. */
	____,____,____,____,____,0012,0027,0033, /* 2. */
	____,____,____,____,____,0005,0006,0007, /* 2. */
	____,____,0026,____,____,____,____,0004, /* 3. */
	____,____,____,____,0024,0025,____,0032, /* 3. */
	 ' ',____,____,____,____,____,____,____, /* 4. */
	____,____, '[', '.', '<', '(', '+', '!', /* 4. */
	 '&',____,____,____,____,____,____,____, /* 5. */
	____,____, ']', '$', '*', ')', ';', '^', /* 5. */
	 '-', '/',____,____,____,____,____,____, /* 6. */
	____,____, '|', ',', '%', '_', '>', '?', /* 6. */ 
	____,____,____,____,____,____,____,____, /* 7. */
	____, '`', ':', '#', '@','\'', '=','\"', /* 7. */
	____, 'a', 'b', 'c', 'd', 'e', 'f', 'g', /* 8. */
	 'h', 'i',____,____,____,____,____,____, /* 8. */
	____, 'j', 'k', 'l', 'm', 'n', 'o', 'p', /* 9. */
	 'q', 'r',____,____,____,____,____,____, /* 9. */
	____, '~', 's', 't', 'u', 'v', 'w', 'x', /* a. */ 
	 'y', 'z',____,____,____,____,____,____, /* a. */
	____,____,____,____,____,____,____,____, /* b. */
	____,____,____,____,____,____,____,____, /* b. */
	 '{', 'A', 'B', 'C', 'D', 'E', 'F', 'G', /* c. */
	 'H', 'I',____,____,____,____,____,____, /* c. */
	 '}', 'J', 'K', 'L', 'M', 'N', 'O', 'P', /* d. */
	 'Q', 'R',____,____,____,____,____,____, /* d. */
	'\\',____, 'S', 'T', 'U', 'V', 'W', 'X', /* e. */
	 'Y', 'Z',____,____,____,____,____,____, /* e. */
	 '0', '1', '2', '3', '4', '5', '6', '7', /* f. */
	 '8', '9',____,____,____,____,____,____  /* f. */
  };
    return (ebc_to_asc[c&0xff]);
}

/*==================================================================*/
#if !BZERO
void bzero(a, n)
/*+++++
.PURPOSE Fill with zeroes
.RETURNS Number of bytes filled
-------*/
  char *a;      /* IN: Starting address */
  size_t  n;	/* IN: Number of bytes to zero */
{
#if 0
  register size_t i ;
  register char *p;

    for (i=n, p=a; i-- > 0; p++)  *p = 0;
#else
    memset(a, 0, n);
#endif
}
#endif 	/* !BZERO */

/*===========================================================================
 *		Quoted String (quote escaped with \)
 *===========================================================================*/

int stkstr(s)
/*++++++++++++++++++
.PURPOSE Match a quoted string
.RETURNS Length of quoted string, INCLUDING terminators
.REMARKS s starts with ' or " or `. The escape character is \ .
---------------------*/
  char	*s;	/* IN: String to scan */
{
  char	*p, quote;
    quote = *s;
    if ((quote != '"') && (quote != '\'') && (quote != '`')) 
  	return(0);
    for (p = s+1; *p; p++) {
  		/* Find End Of String. Take care of quote escaped with \ */
  	if (*p == '\\')	{ p++; continue; }	/* Escaped char	*/
	if (*p == quote){ p++; 	break;	 }	/* Quote found.	*/
    }
    return(p-s);
}

/*==================================================================
		Time utilities
 *==================================================================*/
static time_t t ;
void dateset(adate)
/*++++++++++++++++
.PURPOSE Define the time to edit by date
.RETURNS ---
-----------------*/
  int adate ;	/* IN: Time to use */
{
    t = (time_t)adate ;
}

char *date(op)
/*++++++++++++++++
.PURPOSE  Returns the date in various formats
.RETURNS  The result
.REMARKS  Options are  '.' = fully numeric YYYY.MM.DD,hh:mm:ss
	'D', 'H', 'M', 'S': up to specified level
	'/': 						DD/MM/DD
	'J':						JDdate
	'F': FITS date  YYYY-MM-DD
	'T': FITS date  YYYY-MM-DDThh:mm:ss
-----------------*/
  int op;	/* IN: Option  */
{
  static char buf[28]; 
  struct tm *T;
  char *p, *fmt; 
  double jd;
  int i;

    if (!t) time(&t); 
    switch(op) {
    case '-': case 0:   
    case 'D': case 'd': i = 10; goto case_ctime;
    case 'H': case 'h': i = 13; goto case_ctime;
    case 'M': case 'm': i = 16; goto case_ctime;
    case 'S': case 's': i = 19; goto case_ctime;
    case_ctime:
    	p = ctime(&t); if(p[8]==' ') p[8]='0';
	p[7] = p[24] = 0; p[i] = 0;
    	sprintf(buf, "%s-%s-%s", p+20, p+4, p+8);
    	break;
    case ' ':
    	p = ctime(&t); 
	p[7] = p[10] = p[24] = 0; 
    	sprintf(buf, "%s %s %s", p+20, p+4, p+8);
    	break;
    case '/':
    	T = gmtime(&t);
    	sprintf(buf, "%02d/%02d/%02d", T->tm_mday, T->tm_mon+1, T->tm_year%100);
    	break;
    case 'n': case 'N': case '.':	/* Numeric */
	fmt = "%04d.%02d.%02d,%02d:%02d:%02d" ;
	goto case_numeric ;
    case 'F':				/* FITS	*/
	fmt = "%04d-%02d-%02d" ;
	goto case_numeric ;
    case ':':
    case 'T':				/* FITS */
	fmt = "%04d-%02d-%02d%c%02d:%02d:%02d" ;
	goto case_numeric ;
    case_numeric:
    	T = gmtime(&t); if (T->tm_year < 1000) T->tm_year += 1900;
	sprintf(buf, fmt, T->tm_year, T->tm_mon+1, T->tm_mday, op,
		T->tm_hour, T->tm_min, T->tm_sec);
    	break;
    case 'j': case 'J':			/* JDate   */
	jd = 2440587.5 + (t/86400.);	/* 2440587.5 = JD of Jan. 1, 1970 */
	sprintf(buf, "JD%.5f", jd);
	break;
    default:
    	p = ctime(&t); p[3] = p[7] = p[10] = p[19] = p[24] = 0; 
    	sprintf(buf, "%s %s %s %s, %s", p, p+8, p+4, p+20, p+11);
    	break;
    }
    t = 0;
    return(buf);
}

/*===========================================================================
 *		Find out numbers (equinox + epoch)
 *===========================================================================*/

static int len_epoch(s, opt)
/*++++++++++++++++
.PURPOSE  Find out the length of an equinox/epoch
.RETURNS  Length of string
-----------------*/
  char *s;	/* IN: String to interpret */
  int  opt ;	/* IN: 1 to keep the name  */
{
  char *p;
    if (!s) return(0);
    p = s;
    if (opt) {		/* Keep first word */
	while (isGraph(*p) && (*p != '"') && (*p != ')') && (*p != ']')) p++ ;
	return(p-s);
    }
    if (isalpha(*p)) p += *p == 'A' ? 2 : 1;
    if (*p == ' ') p++;
    while (isdigit(*p)) p++;
    if (*p == '.') for (++p; isdigit(*p); p++) ;
    return(p-s);
}

int isepoch(text)
/*++++++++++++++++
.PURPOSE  Verify the text is an epoch (B or J followed by numbers)
.RETURNS  0(false) / 1(true)
-----------------*/
  char *text;	/* IN: Text to parse */
{
  int i ;
    i = len_epoch(text, 0) ;
    if (i < 4) return(0) ;
    return (! isalnum(text[i])) ;
}

char **epochs(text, opt)
/*++++++++++++++++
.PURPOSE  Find out in a text the Epoch and Equinox
.RETURNS  An array with Equinox (0) and Epoch (1)
.REMARKS  STATIC variables returned. Should be cloned !
	opt=1: accept any name for Epoch ; opt=2: accept any name for Equinox
-----------------*/
  char *text;	/* IN: Text to parse */
  int  opt ;	/* IN: Option 1 to keep the text following Epoch */
{
  static char o[2] = { 2 , 1 } ;
  static char eqp[40];
  static char *val[5] ;		/* Values 2/3 = position of Eq/Ep words	*/
  char *p, *n, *v;
  int typ, len, year, minyear=0;

    val[0] = val[1] = val[2] = val[3] = (char *)0;
    for (p=n=text; *p && (!val[1]); p = n) {
	if (isspace(*n) || iscntrl(*n)) { n++; continue; }
	year = 0;
	for (n=p; isGraph(*n); n++) ;		/* Skip to end of word	*/
	if (*p == '(') p++; if (*p == '[') p++;
	typ = 0 ;				/* Default is Equinox	*/
	if ((p[1] == 'C') && (p[2] == 'R') && (!val[0])) {
						/* Check IRCS System	*/
	    if (strmatch("[HI]CR[SF]*", p)) val[0] = "J2000" ;
	    continue ;
	}
	/* Accept Epoch written as  at "Column-Name" */
	if (val[0] && (strncmp(p, "at \"", 4) == 0)) {
	    p += 4;
	    n = p + strloc(p, '"') ;
	    if (*n) n++ ;
	    v = p ;
	    val[1] = v ;
	    continue ;
	}
	if (toupper(*p) == 'E') { char c;	/* Don't accept EPOCH	*/
	    c = p[1];
	    if (c == 'p') typ = 1;		/* Ep --> Epoch		*/
	    if ((c != 'p') && (c != 'q')) c = 0;
	    else val[typ+2] = p;		/* Word  Epoch/Equinox	*/
	    for (++p; isalpha(*p); p++) ;    	/* Match Epoch/Equinox	*/
	    if (isspace(*p) && c) {		/* Need another word...	*/
		while (isspace(*p)) p++;
		/* Skip the text 'of' , 'is' */
		if ((strncmp(p, "of ", 3) == 0) ||
		    (strncmp(p, "is ", 3) == 0)) {
		    p += 3;
		    while (isspace(*p)) p++;
		}
		n = p;
		while (isGraph(*n)) n++;
	    }
	    while (ispunct(*p)) p++;		/* Skip the ':' or '='	*/
	    if (isspace(*p) && c) {		/* Need another word...	*/
		while (isspace(*p)) p++;
		n = p;
		while (isGraph(*n)) n++;
	    }
	}
	if(val[typ])	continue;		/* Already Found !	*/
	v = p;
	if (opt&o[typ]) {			/* Keep the text..	*/
	    if (*v == '"') v++ ;		/* Romve leading quote	*/
	    val[typ] = v;
	    continue ;
	}
	if ((*p == 'B') || (*p == 'J')) p++;	
	else if ((p[0] == 'A') && (p[1] == 'D')) { /* Accept AD = Anno Domini */
	    p += 2;
	    while (*p == ' ') p++; 
	    v = p;
	    minyear = 100;	  	/* Accept very old year ... */
	}
	else minyear = 1800;
	if (isdigit(*p)) year = atof(p);
	if ((year >= minyear) && (year <= 2050)) {
	    if (isalpha(p[4])) continue;	/* Can be a reference...*/
	    val[typ] = v;
	}
    }

    /* Check if Epoch, but no Equinox */
    if ((!val[0]) && (val[1])) val[0] = val[1] ;

    /* Copy the values in eqp */
    p = eqp ; v = eqp + sizeof(eqp);
    for (typ=0; typ <  4; typ++) {
    	if ((len = len_epoch(val[typ], opt&o[typ]))) {
	    if (len >= (v-p)) len = (v-p) - 1;
	    strncpy(p, val[typ], len);
	    val[typ] = p; p += len;
	    if (p < v) *(p++) = 0;
	}
    }

    return(val) ;
}

double bandGHz(char *text)
/*++++++++++++++++
.PURPOSE  Find the frequency (or wavelength or energy) in text
.RETURNS  The frequency(GHz) / 0 if nothing found
-----------------*/
{
  static double c09, h09;	/* Speed of light / 1.e9 */
  char *p, *w;
  int lw, c;
  double value;

    /* Method: find word ending with 'Hz' 'eV' or 'm' */
    if (!c09) { 
        unit_si("\\c", &value); c09 = value/1.e9; 
        unit_si("2.pi.\\h", &value); h09 = value*1.e9;
    }
    for (w = text; *w; w += lw) {
	for (lw=0; isalnum(w[lw]) || (w[lw] == '.') || issign(w[lw]); lw++) ;
	if (lw==0) {
	    while (strchr(sep, w[lw])) lw++;
	    if (lw==0) return(0);
	    continue;
	}
	if (!isdigit(*w)) continue;
	if (lw<2) continue;
	/* Verify the end: is it a length/frequency/energy ? */
	p = w + lw -2;
	if ( ((p[1] == 'z') && (p[0] == 'H'))  /* Freq */
	  ||  (p[1] == 'm')	               /* Length */
	  || ((p[1] == 'V') && (p[0] == 'e'))) ;  /* E */
	else continue;
	c = w[lw]; w[lw] = 0;
	p = unit_si(w, &value);
#ifdef TEST
        printf("#...bandGHz(%s) => %g (%s)\n", w, value, p);
#endif
	w[lw] = c;
	if (value<=0) continue;
	if (!p) continue;
	if ((p[0] == 'H') && (p[1] == 'z'))
	    return(value/1.e9);
	if ((p[0] == 'm') && (!isalpha(p[1])))
	    return(c09/value);
	if ((p[0] == 'J') && (!isalpha(p[1])))
	    return(value/h09);
    }
    return(0);
}

/*===========================================================================*/
#ifdef TEST
main() 
/*++++++++++++++++
.PURPOSE  Just a TEST program
-----------------*/
{
  static char buf[80], s256[256]; int i, len; char *p, **ep;
  static char timop[] = {'j', 
	'-', '/', 0, 'd', 'h', 'm', 's', '+', ' ', '.', 'a', 'T', 'F'};
  double value;

    for (i=6; --i >= 0; ) printf("----The domain is: %s\n", getdomain()) ;
    gethost(buf, sizeof(buf)) ; 
    printf("----Hostname: %s\n", buf) ;
    for (i=0; i < ITEMS(timop); i++)
    	printf("====date('%c'=%d)=>%s\n", timop[i], timop[i], 
    	date(timop[i]));

    printf("\nLook for bandGHz");
    if (fgets(buf, sizeof(buf), stdin)) while(1) {
	p = s256;
	printf("Your string: "); if (!fgets(p, sizeof(s256), stdin)) break;
	if (! p[0]) break;
	/* Remove newlines */
	for (i=strlen(s256)-1; (i>0)&& isspace(s256[i]); i--) s256[i] = 0;
	value = bandGHz(p);
	printf("====bandGHz(%s) = %g\n", p, value);
    }

    printf("\nLook for Equinox / Epoch: ");
    if (fgets(buf, sizeof(buf), stdin)) while(1) {
	p = s256;
	printf("Your string: "); if (!fgets(p, sizeof(s256), stdin)) break;
	if (! p[0]) break;
	while(1) {
	    p += strlen(p); *(p++) = '\n';
	     printf("(cont) ");
	     if(!fgets(p, sizeof(s256), stdin)) break;
	     if(!*p) break;
	}
	/*ep = epochs(s256);*/
	printf("====Epochs: eq=%s, ep=%s\n", ep[0]?ep[0]:"(nil)",
	ep[1]?ep[1]:"(nil)");
    }


    printf("\nYour template: "); 
    if (fgets(buf, sizeof(buf), stdin)) while(1) {
    	printf("Your string: "); if (!fgets(s256, sizeof(s256), stdin)) break;
    	printf("strmatch(\"%s\", \"%s\") => %d\n", buf, s256,
    		strmatch(buf, s256));
    	printf("stumatch(\"%s\", \"%s\") => %d\n", buf, s256,
    		stumatch(buf, s256));
    }
    
    while(1) {
    	printf("\nYour string: "); if (!fgets(buf, sizeof(buf), stdin)) break;
    	for (i=0; buf[i]; i+=len) {
    	    len = str1(buf+i, s256);
    	    printf("'%s' => %s\n", buf+i, edchar(*s256));
    	}
    	bzero(s256, sizeof(s256));
    	setset1(buf, s256);
    	printf("List of the non-zero bytes:\n");
    	for (i=0; i<sizeof(s256); i++) if (s256[i]) printf("%s,", edchar(i));
    }
}
#endif
