/*++++++++++++++
.IDENTIFICATION xml2fits.c
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT    Unix, CDS Catalogues
.KEYWORDS       Table, FITS, astrores
.VERSION  1.0   07-Mar-2001
.VERSION  1.1   07-Aug-2004: Ignore non-\n-terminated records
.VERSION  1.2   08-Sep-2004: Run in Silent mode
.VERSION  1.3   16-Sep-2004: Add CDS-NAME for the table name
.VERSION  1.4   04-Nov-2004: Don't forget the ending blanks...
.VERSION  1.5   26-Oct-2005: INFO
.VERSION  1.6   20-Dec-2005: No newline in FITS headers!!!
.VERSION  1.7   21-Sep-2007: Use votable
.VERSION  1.8   28-Jul-2008: Problem with referenced tables without name
.VERSION  1.9   19-Aug-2008: Child was not propermy terminated...
.VERSION  1.91  15-Jun-2010: Attribute null= for VALUES
.VERSION  1.92  10-Dec-2010: Last TNULL was not properly taken into account
		Added the usage of a temporary directory.
.COMMENTS	Filter which converts astrores-xml
	(see http://vizier.u-strasbg.fr/doc/astrores.htx)
	or VOTABLE-CSV to FITS
	The output starts at <ASTRO > tag, or <VOTABLE> tag.

    The edition is done in 2 steps:

1. Edit as TSV, which gives in tab_desc e.g.
   #Table  I_284_out
   #Name: I/284/out
   #Title: 280001

   #Coosys J2000_2000.000: eq_FK5 2000_2000.000
   #Field  _1      A24             ID_TARGET
         The original question (from your input file)
   #Field  USNO-B1.0       A12             ID_MAIN
         Designation of the object (1)
          (link)
   #Field  RAJ2000 F10.6   deg     POS_EQ_RA_MAIN
         Right Ascension at Eq=J2000, Ep=J2000 (2)
   #Field  DEJ2000 F10.6   deg     POS_EQ_DEC_MAIN
         Declination at Eq=J2000, Ep=J2000 (2)

2. Transform this into FITS.

   Temporary files (of pattern $TMPDIR/x2f_*) contain the FITS results
   until the actual number of records is known.

   Variable-length char fields are checked and the longest string only
   is kept.

Since version 1.8, a cleanup of the temporary directory is performed,
for files x2f* not having been accessed since TMP_LIFETIME (1800s).

TODO: in case of very long files: resume when there is ~2Gb and start
      a new FITS table ???
---------------*/

#define VERSION "1.92"		/* The Version Number */
#define EMBEDDED_NOTE    0	/* 1 to edit notes as embedded comments \
				   in FITS output			*/

#ifndef DEBUG
#define DEBUG 0
#endif

#ifndef TMP_LIFETIME
#define TMP_LIFETIME 1800
#endif

#include <time.h>
#include <units.h>	/* Definitions of units 	 */
#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>	/* Standard String manipulations */
#include <fcntl.h>	/* File Control definitions	 */
#include <syslog.h>	/* When running by Root		 */
#include <unistd.h>
#include <dirent.h>
#include <ctype.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/stat.h>	/* Find file params  */
#ifndef O_BINARY
#define O_BINARY 0
#endif

#define fit_flushnl fit_flush	/* No newline in FITS headers!!! */
/* Accept 8-bit charactesr */
#define isGraph(c)	(c && (!isspace(c)))

#define MAX(x,y)        ((x) >= (y) ? (x) : (y))
#define MIN(x,y)        ((x) >= (y) ? (y) : (x))
#define ITEMS(a)        sizeof(a)/sizeof(a[0])

/* The input is saved in a buffer */
typedef struct sbuf {
    int size, used, pos;	/* Allocated, used, current position 	*/
    char *text;			/* The actual size depends on size 	*/
} BUF;
#define BUFINCR  (16*1024)	/* Buffer Incrementation by 16k 	*/

#if DEBUG
static char inbyte;
#endif

static char *FITSenv[] = {		/* Environment Variables */
#if O_BINARY		/* MSDOS: Env. variables in UPPERCASE */
    "TOFITSR",		/* Recursivity indicator */
    "TOFITSH",		/* 1 means "write SIMPLE */
    "TOFITSD",		/* 1 means "write data"	 */
    "TOFITSM",		/* 1 means "write FILEMARK */
#else
    "toFITSr",		/* Recursivity indicator */
    "toFITSh",		/* 1 means "write SIMPLE */
    "toFITSd",		/* 1 means "write data"	 */
    "toFITSm",		/* 1 means "write FILEMARK */
#endif
};
#define toFITSr FITSvar[0]
#define toFITSh FITSvar[1]
#define toFITSd FITSvar[2]
#define toFITSm FITSvar[3]
#define starts_by(s,t)	strncmp(s,t,sizeof(t)-1)==0

static char FITSvar[] = { 0, 1, 1, 0 };
static FILE *fout ;		/* Output file		*/
static  int oargc; 		/* Original Arguments	*/
static char **oargv;		/* Orignal arguments	*/

static char *thefile;		/* Input file ----	*/
static char *opened_file;	/* Output file	 	*/
static long maxrecs = -1 ;	/* Maximal output/file	*/
static long saxis2  = -1 ;	/* Position of NAXIS2	*/
/* static long naxis2  = -1 ;	-- Actual      NAXIS2	*/

static char verbop = DEBUG;	/* Verbose option       */
static char ow = 0;  		/* -w options:		*/
#define WOP_fih		2	/* 	-w.fih		*/
#define WOP_fit		4	/* 	-w.fit		*/
static char ofih = 0;  		/* -fih   option	*/
static char ou   = 0;		/* -u option		*/

/* Results are saved in Memory, or to a temporary file datafile */
static char *tmpdir;		/* $TMPDIR */
static char *datafile;		/* Name of temporary data file */
static int   fdata=-1;		/* File handle of temporary file */
static int   vdata=-1;		/* Which kind of table in fdata  */
static int  datarecs ;		/* Number of records written on fdata */
static int  wait4data;		/* How many empty catalogs */
static char *intro_resource;	/* Main header (triggered by RESOURCE)  */
static char *queryparm ;
static char pqueryparm ;	/* !=0 when printed...	*/
static char *fromfile ;		/* =<<==== Query File	*/
static char out_form;		/* m=mini L=list v=Verbose */
static char out_forms[] = "mLv";
static char  concatenate = 0;	/* Concatenate tables ? */
char *sgml_result() ;
int strloc(), setpgm(), error(), gethost();
int sgml_unquote(), sgml_conv(), sgml_fct(), sgml_comment(), sgml_opt();

/* FITS functions used */
int   fit_getmode _ARGS((void));
int   fit_setmode _ARGS((int imode /*(0=standard blocked, 1=stream)*/ ));
int     fit_flush _ARGS((FILE *fout, long offset));
int      fit_prka _ARGS((FILE *fout, char *kw, char *val, char *com));
int      fit_prkn _ARGS((FILE *fout, char *kw, char *val, char *com));
int      fit_prki _ARGS((FILE *fout, char *kw,  int  val, char *com));
int     fit_prcom _ARGS((FILE *fout, char *kw, char *com));
int       fit_prt _ARGS((FILE *fout, char *kw, char *com, int com_len));
int      fit_copy _ARGS((FILE *fout, char *input_file));
int    fit_prfile _ARGS((FILE *fout, char *input_file));
int     fit_naxis _ARGS((int naxis_number));

/* Descriptions are saved in case of Query from File */
typedef struct {
    char *name;
    char *text;
    short nfields;
    short vfields;	/* Number of variable fields */
    int *fields;	/* 2 integers per field: max + variable */
} VTABLE;
static VTABLE *vtable;
static int navot = 0;	/* Number of allocated VTABLEs  */
static int lavot = -1;	/* Number of last allocated one */
static int cavot = -1;	/* Current type of table processed */

static int errors = 0;  /* Errors counter       */
static int clean_pid = 0;

static char usage[] = "\
Usage: xml2fits [-v] [-u[1|2]] [-w[.fit|.fih]] [-noh] [-nodat[a]] [-m max]\n\
       [-o file] [-T tmpdir] [file...]\n\
" ;
static char thelp[] = "\
      -v: verbose option\n\
     -u1: generates as FITS comments the SI values of the units\n\
     -u2: generates both litteral and SI values of the units\n\
    -noh: don't generate the SIMPLE header\n\
  -nodat: generates only the headers (fully blocked to 2880 bytes)\n\
  -m max: specifies a maximal number of data records (default: all)\n\
  -w.fit: generates a new FITS file (extension .fit) for each data file\n\
  -w.fih: generates a header file (extension .fih) for each data file\n\
 -o file: defines the output file (stdout by default)\n\
 -T tdir: directory to store temporary files (default=$TMPDIR or /tmp)\n\
    file: . for no file at all (only the SIMPLE), - for stdin\n\
" ;

static char *info_kw[] = {	/* How to translate INFO */
   "status=CDS-INFO",
   (char *)0
};

/*==================================================================
		Internal Routines
 *==================================================================*/
static void on_death(int s)
/*++++++++++++++++
.PURPOSE  Deal with Zombie
.RETURNS  -
-----------------*/
{
  int st;
    /* signal(SIGCHLD, on_death); */
    if (clean_pid) wait(&st);
    clean_pid = 0;
}

static void terminate(char *text)
/*++++++++++++++++
.PURPOSE  Exit, return to OS
.RETURNS  ---
-----------------*/
{
    if (text) fprintf(stderr, "****xml2fits: %s\n", text);
    if (datafile[0]) {
	close(fdata) ;
	unlink(datafile) ;
    }
    if (clean_pid) {
	kill(clean_pid, SIGTERM);
	on_death(0);
    }
    if (text) exit(1) ;
    exit(0) ;
}

static char *look(char **list, char *kw)
/*++++++++++++++++
.PURPOSE  Find a keyword in a list
.RETURNS  Equivalence
-----------------*/
{
  char **a;
  int i;
    i = strloc(kw, '=');
    for (a=list; *a; a++) {
	if (strncmp(*a, kw, i)) continue;
	if ((!kw[i]) || (kw[i] == a[0][i]))
	    return(*a+i+1);
    }
    return((char *)0);
}

static int iword(char *text, int index)
/*++++++++++++++++
.PURPOSE  Find index of word number #n (count from 0)
.RETURNS  Offset
-----------------*/
{
  char *p ;
    p = text ;
    while (--index >= 0) {
	while(isGraph(*p)) p++ ;
	while(isspace(*p)) p++ ;
    }
    return(p-text) ;
}

static int col_len(char *text)
/*++++++++++++++++
.PURPOSE  Compute length of current column (until \t or \n)
.RETURNS  The length
-----------------*/
{
  char *p;
    for (p=text; *p && (*p != '\t') && (*p != '\n'); p++) ;
    return(p-text) ;
}

static char *col_next(char *text)
/*++++++++++++++++
.PURPOSE  Locate next column (after next tab)
.RETURNS  The next column (points to a TAB or NEWLINE when none)
-----------------*/
{
  char *p;
    for (p=text; *p; p++) {
	if (*p == '\t') return(++p) ;
	if (*p == '\n') return(p); 	/* No next column */
    }
    return(p) ;
}

#if 0
static int buf_append(BUF *b, char *text, int len)
/*++++++++++++++++
.PURPOSE  Append a text in a buffer -- expand if necessary
.RETURNS  0 / -1
-----------------*/
{
  int i, more;
    if (len < 0) len = strlen(text);
    i = b->used + len ;
    if (i >= b->size) { 			/* Must realloc ...	*/
	more = (i|(BUFINCR-1)) + 1 ;
	if (b->text) b->text = realloc(b->text, more) ;
	else  b->text = malloc(more) ;
	b->size = more ;
    }
    memcpy(b->text + b->used, text, len) ;
    b->text[i] = 0;
    b->used = i;
    return(0);
}
#endif

static void open_out(name, ext)
/*++++++++++++++++
.PURPOSE  Open an output file with provided extension
.RETURNS  Opened file / NULL
.REMARKS  Previously opened file is closed;
	The directory is known as cwd.
-----------------*/
  char *name;	/* IN: the data file name ; NULL to close only */
  char *ext;	/* IN: Extension or NULL (see open_ext) */
{
  char *p=0;
    if (fileno(fout) > 2) {
    	fit_flush(fout, -1L);
    	if (ferror(fout)) {
    	    perror(opened_file);
    	    terminate("Error writing FITS output");
    	}
    	fclose(fout);
    	fout = stderr;
    	saxis2 = -1 ;		/* Can't seek..	*/
    }
    if (!name) return;		/* No file to open at all */
    fout = fopen(name, "w+");
    if (!fout) perror(p), terminate("ERROR");
    saxis2 = 0 ;		/* Can   seek..	*/
    opened_file = name;
}

/*==================================================================
		Saved Table Descriptions
 *==================================================================*/
static int saved_vtable(char *name)
/*++++++++++++++++
.PURPOSE  Retrieve the saved description (index in tab_names)
.RETURNS  The index / -1 if not found
.REMARKS  'name' when starting by '#Table' is taken from ref
	Line is e.g.   #Table\tI_254_out\tref=I_254_out\n
-----------------*/
{
  int i, len;
    if (starts_by(name, "#Table\t")) {
	/* Name is given by 'ref=' */
	name = col_next(name);	
	if (*name == '\t') name++;
	if (starts_by(name, "ref=") && isGraph(name[4])) 
	     name += 4;
    }
    for (len=0; isGraph(name[len]); len++) ;

    for (i=lavot; i>=0; i--) {
	if ((strncmp(vtable[i].name, name, len) == 0) 
	         && (vtable[i].name[len] == 0))
	    return(i);
    }
    return(-1);
}

static int save_vtable(char *text)
/*++++++++++++++++
.PURPOSE  Save in tab_descriptions an entry
.RETURNS  The index in tab_names
-----------------*/
{
  VTABLE *pvt;
  char *p, *line;
  int i, j, len, *fields;

    if (saved_vtable(text)<0) {	/* Must add it */
	p = strstr(text, "#Table\t");
	if (!p) {
	    fprintf(stderr, "****Can't be a table: ");
	    len = strlen(text); if(len>40) len=40;
	    fwrite(text, 1, len, stderr);
	    fputc('\n', stderr);
	    return(-1);
	}
	p += 7;
	for (len=0; isGraph(p[len]); len++) ;
        if (++lavot >= navot) {	/* Need reallocation */
	    navot += 256;
	    vtable = vtable ? 
	        (VTABLE *)realloc(vtable, navot*sizeof(VTABLE)) :
	        (VTABLE *)malloc(navot*sizeof(VTABLE));
        }
	pvt = &(vtable[lavot]);
        memset(pvt, 0, sizeof(VTABLE));
        vtable[lavot].text = strdup(text);
        vtable[lavot].name = malloc(len+1); 
	memcpy(vtable[lavot].name, p, len); vtable[lavot].name[len] = 0;
        if (lavot<navot) memset(pvt+1, 0, sizeof(VTABLE));

	/* Count the number of fields (i) + variable-length fields (j) */
	for (i=j=0, line=text; *line; line += len) {
	    len = strloc(line, '\n') ;
	    if (line[len]) len++ ;
	    if (strncmp(line, "#Field\t", 7)) continue ;
	    i++;
	    p = line+iword(line, 2);	/* *p = lettr of format */
	    if (islower(*p)) j++;
	}
	pvt->nfields = i;
	pvt->vfields = j;

	/* Allocate the field structures */
	pvt->fields = fields = (int *)malloc(2*sizeof(int)*(i+1));
	memset(fields, -1, 2*sizeof(int)*(i+1));

	/* Set the various lengths */
	for (line=text; *line; line += len) {
	    len = strloc(line, '\n') ;
	    if (line[len]) len++ ;
	    if (strncmp(line, "#Field\t", 7)) continue ;
	    p = line+iword(line, 2);	/* letter of format */
	    *(fields++) = atoi(p+1);	/* Width of this field */
	    if (islower(*p))		/* Variable-length: actual length */
		*fields = 0;
	    fields++;
	}
    }
    else {
	i = saved_vtable(text);
	if (i<0) return(-1);
	pvt = &(vtable[i]);
    }

    /* Set length of variable fields to zero */
    for (fields = pvt->fields, i=pvt->nfields; --i>=0; fields += 2) {
	if (fields[1]>0) fields[1] = 0;
    }

    return(pvt-vtable);
}

static int reclen(VTABLE *vtable)
/*++++++++++++++++
.PURPOSE  Derive the record length
.RETURNS  record length.
.REMARKS  Don't accept len=0 (replace by len=1)
-----------------*/
{
  int i, len, *f;
    
    if (verbop) fprintf(stderr, "#... reclen(%s)=", vtable->name);
    for (i=len=vtable->nfields, f=vtable->fields; --i>=0; f+=2) {
	if ((f[1]==0) && (f[0]>0)) f[1] = 1;
	if (verbop) fprintf(stderr, " %d", f[1]<0 ? f[0] : f[1]);
	len += f[1]<0 ? f[0] : f[1];
    }
    if (verbop) fprintf(stderr, "\n#   = %d\n", len);
    return(len);
}

/*==================================================================
		Clean-up the directory
 *==================================================================*/

static int cleanup(char *starting_name, time_t last_access)
/*++++++++++++++++
.PURPOSE  Remove the files starting by 'starting_name',
		 not accessed since last_access
.RETURNS  Number of removed files.
-----------------*/
{
  int len, n=0;
  struct dirent *dire;
  struct stat buf;
  DIR *dirp;

    dirp = opendir(".");
    if (!dirp) {
	fprintf(stderr, "****cleanup(./%s*)\n", starting_name);
	exit(1);
    }
    len = strlen(starting_name);
    while ((dire = readdir(dirp))) {
	if (strncmp(dire->d_name, starting_name, len) != 0)
	    continue;
	if (lstat(dire->d_name, &buf) < 0) 	/* Bad entry ??? */
	    continue;
	if (!S_ISREG(buf.st_mode))		/* Not a regular file */
	    continue;
	if (buf.st_atime > last_access)		/* Last access too recent */
	    continue;
	if (verbop) fprintf(stderr, 
		"#...removing old temporary file .../%s\n", dire->d_name);
	if (unlink(dire->d_name)) ;		/* remove file FAILED */
	else n++;
    }
    closedir(dirp);
    return(n);
}

static void clean_launch(char *dirname, int age)
/*++++++++++++++++
.PURPOSE  Remove x2f* files not accessed since "age" seconds
.RETURNS  0 if nothing done
-----------------*/
{
  int len, fh;
  char *fname;
  time_t now;

    now = time(0);
    if (verbop) fprintf(stderr, "....Launching cleanup_dir(%s, %d)\n",
	    dirname, age);
    /* The file /tmp/x2f_0 contains the time of last check.
     * Do this check 6 times during the TMP_LIFETIME
     */
    len = strlen(dirname)+8; if (len<32) len=32;
    fname = malloc(len);
    strcpy(fname, tmpdir); strcat(fname, "/x2f_0");
    fh = open(fname, O_CREAT|O_RDWR, 0666);
    if (fh < 0) {
	fprintf(stderr, "****cleanup_dir(%s): can't read/create ", dirname);
	perror(fname);
	free(fname);
	return;
    }

    len = read(fh, fname, 32);
    if (len>0) {
	fname[len] = 0;
	if (verbop) fprintf(stderr, "....cleanup_dir(%s): %s => age=%ds",
		dirname, fname, (int)now-atoi(fname));
	if(((int)now-atoi(fname)) < age) {
	    if (verbop) fprintf(stderr, " (%s)\n", "too early");
	    close(fh);
	    free(fname);
	    return;
	}
	lseek(fh, 0, 0);
    }
    sprintf(fname, "%u", (unsigned int)now);
    if (verbop) fprintf(stderr, " (%s)\n", "rewrite");
    write(fh, fname, strlen(fname));
    close(fh);
    free(fname); fname = (char *)0;

    /* Start the clean-up in a subprocess */
    signal(SIGCHLD, on_death);
    clean_pid = fork();
    if (clean_pid < 0) {
	fprintf(stderr, "++++fork(pid=%d) failed!\n", getpid());
	clean_pid = 0;
	return;
    }
    if (clean_pid) return;	/* FATHER */

    /* Child: cleanup process to do */
    if (chdir(dirname)) {
	fprintf(stderr, "****Can't chdir ");
	perror(dirname);
	exit(1);
    }
    cleanup("x2f", now-age*6);
    exit(0);
}

/*==================================================================
		FITS-related output
 *==================================================================*/
static int prMARK(name)
/*++++++++++++++++
.PURPOSE  Write out a FILEMARK extension
.RETURNS  Number of lines
-----------------*/
  char *name;	/* IN: Name of next file */
{
  int lines;
    if (verbop) fprintf(stderr, "\t(FILEMARK)\n");
#if DEBUG
    inbyte = getchar();
#endif
    lines = 0;
    lines += fit_prka(fout, "XTENSION", "FILEMARK",
    	"indicates the END of current FITS file");
    lines += fit_prka(fout, "EXTNAME", name, "Name of next FITS file");
    lines += fit_prcom(fout, "END", (char *)0);
    fit_flush(fout, -1L);
    return(lines);
}

static int prSIMPLE(argc, argv)
/*++++++++++++++++
.PURPOSE  Write out the SIMPLE header
.RETURNS  Number of lines
-----------------*/
  int argc;	/* IN: Remaining arguments */
  char **argv;	/* IN: Remaining arguments */
{
static char em[] = "\
         **********************************************************";
  char buf[71], *date(), *getenv(), *p, *e, x;
  int lines;
    if (verbop) fprintf(stderr, "\t(SIMPLE)\n");
    lines = 0;
    e = buf + sizeof(buf) - 1;	/* End of buffer */
#if DEBUG
    inbyte = getchar();
#endif
    lines += fit_prkn(fout, "SIMPLE", "T", "Standard FITS Format");
    lines += fit_prki(fout, "BITPIX", 8, "Character data");
    lines += fit_prki(fout, "NAXIS", 0, "No Image --- just extension(s)");
    lines += fit_prkn(fout, "EXTEND", "T", "There are standard extensions");
    /* lines += fit_prkn(fout, "BLOCKED", "T",
    		"File may be blocked (obsolete, but who knows?)"); */
    sprintf(buf, "xml2fits_v%s", VERSION);
    lines += fit_prka(fout, "ORIGIN", buf,
		"Converted from XML-Astrores to FITS");
    /*
    sprintf(buf, "Converted from XML to FITS (V%s)", VERSION);
    		"File generated at CDS, Strasbourg, France");
    lines += fit_prcom(fout,
    	"                                 (xml2fits, Version %s)", VERSION);
    lines += fit_prcom(fout, buf, "");
    */
    lines += fit_prcom(fout,
    	"                         e-mail: question@simbad.u-strasbg.fr", "");
    strcpy(buf, "ARG=");			/* Write out the Arguments */
    p = buf + strlen(buf); x = '\'';
    while(--argc > 0) {
    	if ((p+strlen(*++argv)) >= &(buf[sizeof(buf)-6])) break;
    	*(p++) = x; x = ' ';
    	strcpy(p, *argv); p += strlen(p);
    }
    *(p++) = '\''; *p = 0;
    if (argc>0) strcpy(p, "...");
    if ((p-buf) > 5) lines += fit_prcom(fout, "COMMENT", buf);
    lines += fit_prka(fout, "LONGSTRN", "OGIP 1.0",
        "Long string convention (&/CONTINUE) may be used") ;
    sprintf(buf, "Written on %s (GMT)", date(':'));	/* Write out the Date */
    lines += fit_prka(fout, "DATE", date('F'), buf);
    p = getenv("USER"); if (!p) p = "(nobody)";
    sprintf(buf, "                             by: %s@", p);
    p = buf + strlen(buf) ;
    gethost(p, e-p) ;
    lines += fit_prcom(fout, buf, "");
    lines += fit_prcom(fout, em, "") ;
    lines += fit_prcom(fout,
	"             EXCERPT from catalogues stored in VizieR (CDS)", "");
    lines += fit_prcom(fout,
	"                     with the following input conditions:", "");
    lines += fit_prcom(fout, em, "") ;
    return(lines);
}

static int prForm(VTABLE *vtable, int records)
/*++++++++++++++++
.PURPOSE  Print out the Table XTENSION Header, from:
	- the table description (stored in tab_desc)
	- the number of records.
.RETURNS  Number of lines wirtten out
.REMARKS  Are set in this routine:
    When records is zero, nothing is written out -- only
	the various lengths in vtable are computed.
-----------------*/
{
  char *p, *e, *f, *t, *u, *n, *line, x, y, z, buf[256], kbuf[32];
  int len, i, j, m, w, o, *field;
  int lines = 0;

    /* Records not specified ===> Just return */
    /* if (records == 0) return(lines) ; */

    /* Records are known -- Edit the full XTENSION header */
    lines += fit_prka(fout, "XTENSION", "TABLE",
	"Ascii Table Extension (TAB and NEWLINE sep)");
    lines += fit_prki(fout, "BITPIX", 8, "Character data");
    lines += fit_prki(fout, "NAXIS", 2, "Simple 2-D matrix");
    lines += fit_prki(fout, "NAXIS1", 1+reclen(vtable), 
	    "Number of bytes per record");
    lines += fit_prki(fout, "NAXIS2", records, "Number of records");
    lines += fit_prki(fout, "PCOUNT", 0, "Get rid of random parameters");
    lines += fit_prki(fout, "GCOUNT", 1, "Only one group (isn't it obvious?)");

    lines += fit_prki(fout, "TFIELDS", vtable->nfields, 
	    "Number of data fields (columns)");

    /* Edit the Catalog Details */
    n = t = (char *)0 ;				/* <INFO> details	*/
    for (p=intro_resource; *p; p += len) {
	len = strloc(p, '\n') ;
	x = p[len] ; p[len] = 0 ;
	if (*p == '#') t = p+iword(p,1) ;	/* 1st Word after #xxx	*/
	if (strncmp(p, "#Name:", 6) == 0) {
	    if (wait4data) --wait4data, 	/* Catalogue Empty	*/
		 fit_prka(fout, (char *)0, t,
		 "Catalogue skipped (nothing found there!)") ;
	    else fit_prka(fout, "CDS-CAT", t,
		"Catalogue designation in CDS nomenclature");
	    lines++;
	}
	else if (strncmp(p, "#Title:", 7) == 0) {
	    fit_prcom(fout, "       ", t) ;
	    lines++;
	}
	else if (strncmp(p, "#INFO\t", 6) == 0) {	/* <INFO> */
	    f = p+6; while (*f == ' ') f++;
	    n = look(info_kw, f);
	    if (n) {
		f += strloc(f, '=');	/* f->value to associate to keyword */
		if (f[0] == '=') f++;
		i = strloc(f, '\t');
		t = f + i;		/* t->comment */
		if (*t) t++;
		for (--i; isspace(f[i]); i--) ; ++i;
		y = f[i]; f[i] = 0;
		j = strlen(t);
		if (j>0) { if (t[--j] != '#') j++; }
		if (j>0) { for (--j; isspace(t[j]); j--) ; ++j; }
		z = t[j]; t[j] = 0;
		lines += fit_prka(fout, n, f, t);
		t[j] = z;
		f[i] = y;
	    }
	}
	if ((p[len] = x)) len++;
    }

    /* Edit now the Format */
    o = 1;	/* V1.6: Keep offset#0 for newline */
    m = vtable->nfields;
    field = vtable->fields;
    for (j=0, line=vtable->text; *line; line += len) {
        len = strloc(line, '\n'); if (line[len]) len++;
        if (strncmp(line, "#Table\t", 7) == 0) {
    	    lines += fit_prka(fout, "EXTNAME", vtable->name,
		"Identification of the table");
	    continue ;
	}
	if (strncmp(line, "#Name:", 6) == 0) {	/* Name of Table */
	    for (t=line+6; isspace(*t); t++);
	    for (p=t; isgraph(*p); p++) ;
	    x = *p; *p = 0;
	    lines += fit_prka(fout, "CDS-NAME", t, "Table name in METAtab");
	    *p = x;
	    continue;
	}
	if (strncmp(line, "#Title:", 7) == 0) {
	    for (p=buf, i=10; --i>=0; p++) *p = ' ' ;
	    strncpy(buf+10, line+7, sizeof(buf)-10) ;
	    buf[strloc(buf, '\n')] = 0 ;
	    fit_prcom(fout, "       ", buf) ;
	    continue ;
	}
	if (strncmp(line, "#Values(", 8) == 0) { /* V1.91: add TNULL */
	    n = col_next(line) ;       /* NULL value */
	    if (iscntrl(*n)) continue; /* TAB not found */
	    sprintf(kbuf, "TNULL%d", j);
	    for (i=0; n[i] && (!iscntrl(n[i])); i++); x = n[i]; n[i] = 0;
	    lines += fit_prka(fout, kbuf, n, "NULL definition");
	    n[i] = x;
	}
	if (strncmp(line, "#Field\t", 7)) continue ;

	/* Width of Column */
	w = field[1]<0 ? field[0] : field[1]; 
	n = col_next(line) ;	/* Name of Column */
	f = col_next(n) ;  	/* Format */
	u = col_next(f) ; 	/* Unit   */
	t = col_next(u) ;	/* UCD	  */

	++j ;
        sprintf(kbuf, "TBCOL%d", j);
	i = strloc(t, '\n'); if (i>30) i = 30; x = t[i]; t[i] = 0 ;
        sprintf(buf, "UCD=%s", t) ;
	t[i] = x ;
	i = strlen(buf); while (i < 30) buf[i++] = '.' ;
        sprintf(buf+i, " offset=%d", o) ;
        lines += fit_prki(fout, kbuf, o+1, buf) ;
#if 0
	if (j == _1pos) {	/* For the _1 Column: reformat needed */
	    sprintf(buf, "A%d", w);
	    f = buf;
	}
	sprintf(kbuf, "TFORM%d", j);
	i = strloc(f, '\t'); x = f[i]; f[i] = 0 ;
	y = f[0]; f[0] = toupper(f[0]); 
	lines += fit_prka(fout, kbuf, f, "Fortran Format");
	f[i] = x ; f[0] = y;
#else
	sprintf(kbuf, "TFORM%d", j);
	if (field[1]<0) {	/* Non-variable-length field */
	    i = strloc(f, '\t'); if (i>30) i = 30;
	    strncpy(buf, f, i); buf[i] = 0;
	}
	else 			/* Variable-length field */
	    sprintf(buf, " %d", w);
	buf[0] = toupper(f[0]);
	lines += fit_prka(fout, kbuf, buf, "Fortran Format");
#endif
	o += w ;
	field += 2;

	/* TTYPE = Label */
	sprintf(kbuf, "TTYPE%d", j);
	i = strloc(n, '\t'); x = n[i]; n[i] = 0 ;
	t = line + len ;	/* Explanations */
	y = 0 ;
	if (*t && (*t != '#')) {
	    while ((*t == ' ') || (*t == '\t')) t++ ;
	    e = t + strloc(t, '\n') ;
	    len = e - line ;	/* Index of \n */
	    y = line[len]; line[len] = 0 ;
	}
	else t = "Label of Column" ;
	lines += fit_prka(fout, kbuf, n, t);
	n[i] = x ;
	if (y) line[len] = y ;

	/* Units */
	if (isGraph(*u)) {
	    i = strloc(u, '\t'); x = u[i]; u[i] = 0 ;
            t = unitex(u, ou);
            if (*t) {       /* Units */
	        sprintf(kbuf, "TUNIT%d", j);
	        lines += fit_prka(fout, kbuf, u, t);
	    }
	    u[i] = x ;
	}
	o++ ;
    }
#if 0
    /* Keep the Description when Query from a File */
    if (fromfile && tabID[0] && tab_desc && (!datafile[0])) {
	save_vtable(vtable->name, tab_desc) ;
    }
#endif
    return(lines) ;
}

static int prFile(VTABLE *pvt, int fdata)
/*++++++++++++++++
.PURPOSE  Transform the data file to FITS
.RETURNS  Number of records
-----------------*/
{
  int c, j=0, o, lines=0, w=0, stat, *f=(int *)0;
  char buf[BUFSIZ], *b, *e;
  long bytes=0;		/* Number of bytes issued */

    lseek(fdata, 0, 0);

    /* fprintf(stderr, "....DATA output....\n%s\n", data) ; */
    /* Here b = current byte from buffer,
     *      c = current character from input,
     *      o = current position in output line (o=0 for beginning of line)
     *      j = current position in current column
     *      w = width of current column
     */
    b = e = buf;		/* Input buffer ... */
    o = 0;			/* Offset in output */
    stat = 1;			/* Reading status.. */
    while (stat>0) {
	if (b>=e) {		/* Read next buffer */
	    stat = read(fdata, b=buf, sizeof(buf));
	    e = b+stat;
	    continue;
	}
        if (o == 0) { 		/* Beginning of a line */
	    fputc('\n', fout); o++; bytes++; 
	    f = pvt->fields;
	    w = f[1]<0 ? f[0] : f[1];
	    j = 0;
	    lines++;
	}
	c = *(b++);
	if ((c == '\t') || (c == '\n')) {
	    /* Terminate current column */
	    while (j<w) { fputc(' ', fout); j++; o++; bytes++; }
	    j = 0;
	    if (f[2]>=0) {	/* Another column follows   */
		f += 2;
		w = f[1]<0 ? f[0] : f[1];
		fputc('\t', fout); o++; bytes++;
		if (c == '\n') --b;
		continue;
	    }
	    /* Should be the last column ! */
	    if (c == '\t') {	/* An extra column in the input ? */
		w = 0;
		continue;
	    }
	    fputc('\r', fout); bytes++;
	    o = 0;		/* Now at a line beginning */
	    continue;
	}
	/* Write out current text */
	if (j<w) {		/* Truncate too large column... */
	    fputc(c, fout); 
	    j++, o++, bytes++;
	}
    }
    fit_flushnl(fout, bytes); 	/* Fill the last block with blanks */
    if (stat<0) {		/* Error... */
	fprintf(stderr, "****xml2fits: ");
	perror("reading input (xml)");
    }

    return(lines);
}

static int prData(VTABLE *pvt, char *data)
/*++++++++++++++++
.PURPOSE  Print out the Data file
.RETURNS  Number of data lines
.REMARKS  reclen contains the starting \n, but not the ending \r
-----------------*/
{
  int i, j, len, lines=0, w, *f;
  char *p, *s, tab;
  long bytes=0;		/* Number of bytes issued */

    /* fprintf(stderr, "....DATA output....\n%s\n", data) ; */
    for (p = data ; *p; p += len) {
        len = strloc(p, '\n') ;
	if (p[len]) len++ ;
	tab = '\n';			/* Leftmost character */
	f = pvt->fields;
	for (s=p, i=pvt->nfields; --i>=0; f+=2, s=col_next(s)) {
	    fputc(tab, fout); bytes++; tab = '\t';
	    j = col_len(s);
	    w = f[1]<0 ? f[0] : f[1];
	    if (j>w) j = w;		/* Truncate if too large */
    	    fwrite(s, 1, j, fout); 
	    while (j<w) { fputc(' ', fout); j++; }
	    bytes += w;
	}
	fputc('\r', fout); bytes++;	/* Ending the line... */
	lines++;
    }
    fit_flushnl(fout, bytes); 	/* Fill the last block with blanks */
    return(lines);
}

static int prTitle(text)
/*++++++++++++++++
.PURPOSE  Print out the Title (triggered by <TABLE>)
.RETURNS  Number of lines
.REMARKS  Also set the pecularities
	fromfile  (result of an input from a file)
	out_form  (mini / verbose output)
-----------------*/
  char *text;	/* IN: Text of title including number, author, etc */
{
  char *p, x ;
  int i, len, lines=0 ;

#if DEBUG>1
    printf("prTitle: %s", text); inbyte = getchar();
#endif
    lines = 0;

    if (queryparm) for (p=queryparm; *p; p += len)  {
	if (*p == '\n') p++ ;
	len = strloc(p, '\n') ;
	if (*p != ' ') {		/* Not input param.? */
	    if (!fromfile) continue ;
	}
	if (*p == ' ') {
	    p++, len-- ;
	    if (len == 0) continue ;
	}
	i = strloc(p, '=') ;
	if (i >= len) {			/* The end...        */
	    if (!fromfile) continue ;
	}
	if (i ==(len-1)) continue ;	/* Useless parameter */
	if (*p == '!') continue ;
	/* Ignore -to, -from, -this */
	if (*p == '-') {
	    if (starts_by(p, "-from=")) continue ;
	    if (starts_by(p, "-to=")) continue ;
	    if (starts_by(p, "-this=")) continue ;
	    if (starts_by(p, "-out.form=") && (!out_form)) {
		/* Accept mini List (or Titles) and verb */
		if (strchr(out_forms, p[10])) out_form = p[10];
	    }
# if 0
	    if (starts_by(p, "-out.add=_1") && (!datafile[0])) {
		/* Use a temporary file */
		sprintf(datafile, "%s/x2f_%d", tmpdir, getpid()) ;
		fdata = open(datafile, O_CREAT|O_RDWR|O_TRUNC, 0666) ;
		if (fdata < 0) {
		    fprintf(stderr, "****") ;
		    perror(datafile) ;
		    exit(1) ;
		}
	    }
#endif
	}
	else if (starts_by(p, "#out.form=") && (!out_form)) {
	    /* Accept mini List (or Titles) and verb */
	    if (strchr(out_forms, p[10])) out_form = p[10];
	}
	x = p[len]; p[len] = 0 ;
	if (!fromfile) {	/* Set flag for Query from File */
	    if (strncmp(p+i, "=<<====", 7) == 0)
		fromfile = strdup(p) ;
	}
	fit_prcom(fout, "", p) ;
	p[len] = x ;
    }
    if (queryparm) pqueryparm = 1 ;

    return(lines);
}

static int write_datafile()
/*++++++++++++++++
.PURPOSE  Write the data file (even when no records)
.RETURNS  Number of records written out
-----------------*/
{
  int nrecs;

    prForm(vtable+vdata, datarecs);
    fit_prcom(fout, "END", (char *)0); 	/* Write the "End" */
    fit_flushnl(fout, -1L);
    nrecs = prFile(vtable+vdata, fdata);

    /* Reset variables */
    lseek(fdata, 0, 0);
    datarecs = 0;
    vdata = -1;

    return(nrecs);
}

#if 0
static void setup()
/*++++++++++++++++
.PURPOSE  Set up what's required for a new file.
.RETURNS  ---
-----------------*/
{
  int i ;

    toFITSr += 1;			/* Recursivity counter	*/
    if (ow&WOP_fit) toFITSd = 1;
    if (ow&WOP_fih) toFITSd = 0;
    if (toFITSm) toFITSh = 0;	/* If FILEMARK no initial SIMPLE */
    fit_setmode(toFITSd);	/* With option "-nodata", simple ASCII file */
    if (verbop) {
	for (i=0; i<ITEMS(FITSenv); i++) fprintf(stderr,
	     "....Env: %s=%d\n", FITSenv[i], FITSvar[i]);
    }

    if (toFITSr > 1) return ;		/* Setup executed once.	*/

}
#endif

char *commesave(char *text)
/*++++++++++++++++
.PURPOSE  Get the Comments -- when these contain the Query Parameters.
.RETURNS  ---
.REMARKS  The comments contain the &lt; etc -- to be changed here...
-----------------*/
{
  char *p = (char *)0;
    if (queryparm && (!datafile[0])) {
	if (fromfile && strstr(text, "Next constraint")) {
	    p = text ;
	    free(queryparm);
	}
	else return("") ;
    }
    if (!p) {
        for (p=text; isspace(*p); p++) ;
        if (!starts_by(p, "Parameters used for the Query"))
	    p = (char *)0 ;
    }
    if (p) {
        queryparm = strdup(p) ;
	sgml_unquote(queryparm) ;
    }
    return("") ;
}

char *tabastart() {
/*++++++++++++++++
.PURPOSE  Triggered by <TABLE>
.RETURNS  ""
-----------------*/
  static int counts = 0 ;
  char *p ;
    p = sgml_result() ;
    while (isspace(*p)) p++ ;
    if (intro_resource)  {	/* Concatenate when -out.add=_1 */
        free(intro_resource);
    }
    intro_resource = strdup(p) ;
    if (verbop) fprintf(stderr, "....tabastart:\n%s\n", intro_resource) ;
    if (counts == 0) {
	prTitle(intro_resource);
        fit_prcom(fout, "END", (char *)0); 	/* Write the "End" */
        fit_flushnl(fout, -1L) ;
    }
    counts++ ;
    wait4data++ ;			/* Decremented in endata   */
    return("") ;
}

char *datastart() {
/*++++++++++++++++
.PURPOSE  Triggered by <CSV>
.RETURNS  ""
-----------------*/
  static int tabno = 0;
  char *p ;

    p = sgml_result() ;		/* Contains all field definitions */
    while (isspace(*p)) p++ ;
    /* Should we try to concatenate data ? */
    if ((tabno==0) && (out_form == 'm') && fromfile) {
	concatenate = 1;
    }

#if 0
    /* Is the table already known ? */
    if ((tabno>0) && concatenate) {
	if (saved_vtable(p) != vdata) {	/* It's not the same table! */
	    if (datarecs>0) write_datafile();
	    vdata = -1;
	}
    }
#endif

    /* Save the tabular description */
    cavot = save_vtable(p);
    if ((tabno==0) && concatenate) {
	/* Use a temporary file */
	sprintf(datafile, "%s/x2f_%d", tmpdir, getpid()) ;
	fdata = open(datafile, O_CREAT|O_RDWR|O_TRUNC, 0666) ;
	if (fdata < 0) {
	    fprintf(stderr, "****") ;
	    perror(datafile) ;
	    exit(1) ;
	}
	if (verbop) fprintf(stderr, 
		"...use temporary file: %s\n", datafile);
	vdata = cavot;
    }

    tabno++;
    if (verbop) fprintf(stderr, "....datastart #%d (%s):\n%s\n", 
	    tabno,
	    "VARIABLE dimension is lowercase",
	    vtable[cavot].text) ;
    return("") ;
}

char *endata() {
/*++++++++++++++++
.PURPOSE  Triggered by </CSV>
.RETURNS  ""
-----------------*/
  VTABLE *pvt;
  int nrecs = 0;
  int len, i, colen, headlines=0, *f;
  char *p, *s, *e, *adata, *edata;

    p = sgml_result() ;
    while (isspace(*p)) p++ ;
    adata = p ;
    /* Ignore non-terminated records */
    for (p=adata+strlen(adata)-1; (p>=adata) && (*p != '\n'); p--) 
	*p = 0;
    edata = p;
    wait4data-- ;
    pvt = vtable+cavot;

    /* Count the Number of Records */
    if (adata && *adata) {
	if (*adata == '#') {
	    headlines = atoi(adata+iword(adata,1));
	    adata += strloc(adata, '\n') ;
	    if (*adata) adata++ ;
	}
	/* Skip Headlines */
	while (--headlines >= 0) {
	    adata += strloc(adata, '\n') ;
	    if (*adata) adata++ ;
	}
	/* Skip empty lines */
	while (*adata == '\n') adata++;
	/* Compute the Number of Records and length of variable-length fields */
	for (p=adata; p<edata; p += len) {
	    len = strloc(p, '\n') ;
	    if (p[len]) nrecs++, len++ ;
	    if (pvt->vfields) {	/* Compute max. length of variable-length */
		f = pvt->fields;
		for (s=p, i=pvt->nfields; --i>0; f+=2, s=col_next(s)) {
		    if (f[1]<0) continue;	/* Fixed-length column */
		    for (e=s+col_len(s); (e>=s) && isspace(*e); e--) ;
		    colen = (++e) - s;
		    if (colen>f[1]) f[1] = colen;
		}
	    }
	}
    }

    /* Nothing found and mini ==> return */
    if ((nrecs==0) && (out_form == 'm'))
	return("");

    /* When the records are saved to a file... */
    if (concatenate) {
	if (vdata != cavot) {	/* The file contains another type of table */
	    if (datarecs>0) write_datafile();
	    vdata = cavot;
	}
	write(fdata, adata, 1+(edata-adata));	/* 1+ to write the newline */
	datarecs += nrecs ;
	return("");
    }
#if 0
    if (datafile[0]) {
	_1len = len_1;		/* Save length of _1 column for prFile */
	datarecs += nrecs ;
	while (*adata) {
	    len = strloc(p = adata, '\n') ;
	    if (!adata[len]) break;
	    if (_1len) { 	/* Find the _1 column */
	        i = _1pos;
	        while (--i > 0) {
		    p += strloc(p, '\t') ;
		    if (*p) p++ ;
	        }
		if (p > adata) 	/* Write what's BEFORE the _1 column */
		    write(fdata, adata, p-adata) ;
	        i = col_len(p);
	        n = i - _1len ;
	        if (n < 0) {	/* Must insert blanks */
		    write(fdata, p, i) ;
		    while (n<0) write(fdata, " ", 1), n++ ;
	        }
	        else {
		    write(fdata, p, _1len);
	        }
		p += i;					/* At next tab */
	    }
	    if (adata[len]) len++;
	    adata += len ;
	    write(fdata, p, adata-p) ;
	}
	if ((nrecs) && (adata[-1] != '\n')) write(fdata, "\n", 1) ;
	return("") ;
    }
#endif

    /* Issue the "XTENSION" for ascii tables;
	either from a ".fih" file, or via prForm.
	The variable "i" is set to 1 if we've to
	use such a .fih file.
    */
    if (ow&WOP_fih) open_out(thefile, ".fih");
    i = prForm(pvt, nrecs);
    i += fit_prcom(fout, "END", (char *)0); 	/* Write the "End" */
    fit_flushnl(fout, -1L);

    /* Now, issue the Data ... */
    if ((toFITSd==1) && maxrecs)
	prData(pvt, adata) ;

    return("") ;
}

static void convert(char *file)
/*++++++++++++++++
.PURPOSE  Convert a file into FITS
.RETURNS  ---
-----------------*/
{
  static int Ncalls = 0;
  char *include, *old_file, *sgml_exec() ;

    /* Initialisation: be sure the temporary file name can be found */
    if (!tmpdir) tmpdir = getenv("TMPDIR");
    if (!tmpdir) tmpdir = "/tmp";
    if (!datafile) datafile = malloc((strlen(tmpdir)|7)+17);

    /* Cleanup directory at initial call only */
    if (Ncalls == 0) clean_launch(tmpdir, TMP_LIFETIME/6);
    Ncalls++;

    old_file = thefile; thefile = file ;
    if (file) {
	include = malloc(strlen(file)+16) ;
	sprintf(include, "${.include %s}", file) ;
    }
    else {
	file = "(stdin)" ;
	include = (char *)0 ;
    }
    if (verbop) fprintf(stderr, "....Dealing with file: %s\n", file);

    /* Write first the FILEMARK if necessary */
    if (toFITSm) {
	if (!ow) prMARK(file);
	toFITSm = 0;
	toFITSh = 1;	/* Must be followed by a SIMPLE header */
    }

    /* Open the Output FITS File if Necessary */
    if (ow&WOP_fit) {
    	toFITSh = 1;
	open_out(file, ".fit");
    }

    /* Write out the SIMPLE header (except if a FITS file
    	is to be issued, done previously)
    */
    if (toFITSh) {
	if (ow&WOP_fih) open_out("SIMPLE", ".fih");
    	prSIMPLE(oargc, oargv);
    	toFITSh = 0;
    }

    /* Execute the conversion, with result saved */
    sgml_conv(include) ;
    if (include) free(include) ;
    /* result = sgml_exec((char *)0) ; */

    /* Terminate the File */
    if (datarecs>0) {
	write_datafile();
    }
    else if (queryparm && (!pqueryparm)) {
	prTitle((char *)0) ;
	fit_prcom(fout, "HISTORY", "**** No Matching Record ****") ;
	fit_prcom(fout, "       ", "**** No Matching Record ****") ;
	fit_prcom(fout, "COMMENT", "**** No Matching Record ****") ;
        fit_prcom(fout, "END", (char *)0); 	/* Write the "End" */
        fit_flushnl(fout, -1L);
    }

    thefile = old_file ;
}


/*==================================================================*/

#define pVal(text)      \
	if (! *++p) {                           	\
	    if (--argc <= 0) errors++, error(text);   	\
	    else    p = *++argv;            		\
	}

/* ================================================================
<!--
#### Definitions of SGML useful tags in Astrores (XML)
#### for transformation into FITS.

NOTE: The following basic keys are used:
${}          = name of current tag
${^}         = be sure to be at left margin
${...}       = get input until end of tag
${.ignore}   = just forget what comes (until end-of-tag)
${.save}     = save in specified variable (until end-of-tag)
----------------------------------------------------------->
=================================================================== */
static char fits_defs[] = "\
<DEF SORTED -- Sorted (case insensitive) Definitions>\n\
&Epsi;={Epsilon}\n\
&Upsi;={Upsilon}\n\
&angl0;=<\n\
&angr0;=>\n\
&angst;={AA}\n\
&ap;={approx}\n\
&arcmin;='\n\
&arcsec;=\"\n\
&d2;={dot}\n\
&d3;={ddot}\n\
&epsi;={epsilon}\n\
&epsis;={epsilon}\n\
&farcm;='.\n\
&farcs;=\".\n\
&fdg;={deg}.\n\
&fm;=^m^.\n\
&j0;=o\n\
&larr;=<--\n\
&odot;={sun}\n\
&parl0;(\n\
&parr0;)\n\
&phis;={phi}\n\
&phiv;={phi}\n\
&prime;='\n\
&prop;={prop.to}\n\
&rarr;=-->\n\
&sime;==~\n\
&solm0;=M\n\
&sqbl0;=[\n\
&sqbr0;=]\n\
&thetas;={theta}\n\
&thetav;={theta}\n\
&thinsp;=\n\
&upsi;={upsilon}\n\
&vbm0;=|\n\
<ASTRO>=${^}#${...}\n\
<COOSYS>=${^}#Coosys\t${ID}:\t${system} ${equinox} ${...}\n\
<CSV>=${.datastart}${^}#Data ${headlines}${.nl2cr}${...}${.endata}\n\
<DATA>=${^}${...}${\\n}\n\
<DEFINITIONS>=${^}#${...}\n\
<DESCRIPTION>=${.obeyline #}${...}\n\
<FIELD>=${^}#Field\t${name}\t${.fitsDISP}\t${@unit}\t${UCD}${...}\n\
<INFO>=${^}#INFO\t${name}=${value}\t${...}\t#\n\
<LINK>= (link)\n\
<NAME>=${^}#Name: ${...}\n\
<RESOURCE>=${^}${\\n}#RESOURCE=${ID}${\\n}#Name: ${name}${\\n}\
#Title: ${.gobble}${...}\n\
<TABLE>=${.tabastart}${^}#Table\t${ID}\tref=${ref}${\\n}#Name: ${name}${\\n}\
#Title: ${.gobble}${...}\n\
<TITLE>=${^}#Title:\t${...}\n\
<VALUES>=${^}#Values(${type}):\t${null}\tlist:${\\n}#${...}\n\
<VOTABLE>=${^}#${...}\n\
</DEF>" ;
/*<VALUES>=${^}#---List of ${type} values:${...}\n\  [mod. V1.91]*/

/*==================================================================
		Program
 *==================================================================*/
int main(int argc, char **argv)
/*++++++++++++++++
.PURPOSE  Program which concerts to FITS
-----------------*/
{
  char *p, *tparm, *getenv(), *getcwd();
  int i, nf;

    setpgm("xml2fits");
    if (!fout) fout = stdout ;
    oargc = argc; oargv = argv;
    nf = 0 ;	/* Number of files processed */
    tparm = (char *)0 ;

    /* "name" is required -- set it to "" */
    putenv("name=") ;

    /* Under ftpd, redirect stderr to 1 */
    /* if (getuid() == 12) { close(2); dup(1) ; } */
    /* fprintf(stderr, "sizeof(cwd)=%d\n", sizeof(cwd)); */

    /* Check first the Environment */
    for (i=0; i<ITEMS(FITSenv); i++) {
    	if ((p = getenv(FITSenv[i]))) FITSvar[i] = atoi(p);
    }
    /* prSIMPLE(oargc, oargv) ; */

    /* Insert the Dictionary for FITS */
    sgml_conv(fits_defs) ;
    sgml_fct("${.datastart}", datastart) ;
    sgml_fct("${.tabastart}", tabastart) ;
    sgml_fct("${.endata}", endata) ;
    sgml_comment(commesave) ;

    /* Loop on the Arguments */
    sgml_opt("-s");		/* V1.2: Be sure the 'silent' option is on ! */
    while (-- argc > 0) {
	p = *++argv;
	if (*p == '-') switch (*++p) {
	  case 'o' :                   	/* Tell the output file	*/
	  	pVal("Missing -o(utput) file");
		fflush(fout) ;
		open_out(p, (char *)0); /* This closes current	*/
	  	continue;
	  case 'h' :                   	/* HELP */
		printf("%s%s", usage, thelp) ;
		exit(0) ;
	  case 'm' :                   	/* Maximum records...	*/
		pVal("Missing -m(ax_records) value");
		maxrecs = atol(p) ;
    		continue ;
	  case 'v':                     /* Verbose /version     */
		if (strncmp(p, "ver", 3) == 0) {
		    printf("xml2fits (CDS) Version %s\n", VERSION);
		    exit(0);
		}
		verbop = 1;
		sgml_opt(*argv);
		fprintf(stderr, "....PATH=%s\n", getenv("PATH"));
		/* osmtrace(1); */
		continue;
	  case 'f':			/* -fih (generate from *.fih) 	*/
	  	if (strcmp(p, "fih") == 0) {
	  	    ofih = 1;
	  	    continue;
	    	}
	  	goto bad_option;
	  case 'n':			/* -nodata	*/
	  	if (strcmp(p, "nodata") == 0) { toFITSd = 0; continue; }
	  	if (strcmp(p, "nodat") == 0)  { toFITSd = 2; continue; }
	  	if (strcmp(p, "noh") == 0)    { toFITSh = 0; continue; }
	  	goto bad_option;
	  case 'u':			/* -u		*/
		ou = atoi(p+1);
		continue;
	  case 'w':			/* -w		*/
	  	if (strcmp(p, "w.fit") == 0) 	   ow |= 1|WOP_fit;
	  	else if (strcmp(p, "w"    ) == 0) ow |= 1|WOP_fit;
	  	else if (strcmp(p, "w.fih") == 0) ow |= 1|WOP_fih;
	  	else goto bad_option;
    		if ((ow &(WOP_fit|WOP_fih)) == (WOP_fit|WOP_fih))
    		    terminate("Can't have options  -w   and  -w.fih");
	  	continue ;
	  case 'T':			/* Temporary directory  */
		pVal("Missing -T(empdir) filename");
		tmpdir = p;
		continue;
	  case '\0':			/* - alone		*/
    	    	if (tparm) convert(tparm) ;
		nf++ ;			/* Don't process tables	*/
		continue ;
	  default:
	  bad_option:
		fprintf(stderr, "\n****Bad option: -%s\n", p);
		errors++;
		terminate("ERROR");
	}

	/*----------- A file to Process ---------------------------------*/

	/* A file to process -- set it as an ${.include} */
	convert(p) ;
	nf++ ;
    }

    if (nf  == 0) convert((char*)0);
    terminate((char *)0);
    exit(0);		/* Useless, but the compiler requires it... */
}
